package cn.com.qmth.examcloud.core.examwork.service.impl;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

import cn.com.qmth.examcloud.api.commons.enums.CURD;
import cn.com.qmth.examcloud.api.commons.enums.ExamType;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.DynamicEnum;
import cn.com.qmth.examcloud.commons.helpers.DynamicEnumManager;
import cn.com.qmth.examcloud.commons.helpers.poi.ExcelReader;
import cn.com.qmth.examcloud.commons.util.DateUtil;
import cn.com.qmth.examcloud.commons.util.DateUtil.DatePatterns;
import cn.com.qmth.examcloud.commons.util.PathUtil;
import cn.com.qmth.examcloud.core.basic.api.OrgCloudService;
import cn.com.qmth.examcloud.core.basic.api.bean.OrgBean;
import cn.com.qmth.examcloud.core.basic.api.request.GetOrgReq;
import cn.com.qmth.examcloud.core.basic.api.response.GetOrgResp;
import cn.com.qmth.examcloud.core.examwork.base.enums.ExamProperty;
import cn.com.qmth.examcloud.core.examwork.dao.ExamOrgPropertyRepo;
import cn.com.qmth.examcloud.core.examwork.dao.ExamPropertyRepo;
import cn.com.qmth.examcloud.core.examwork.dao.ExamRepo;
import cn.com.qmth.examcloud.core.examwork.dao.ExamSpecialSettingsRepo;
import cn.com.qmth.examcloud.core.examwork.dao.entity.ExamEntity;
import cn.com.qmth.examcloud.core.examwork.dao.entity.ExamOrgPropertyEntity;
import cn.com.qmth.examcloud.core.examwork.dao.entity.ExamPropertyEntity;
import cn.com.qmth.examcloud.core.examwork.dao.entity.ExamSpecialSettingsEntity;
import cn.com.qmth.examcloud.core.examwork.service.ExamService;
import cn.com.qmth.examcloud.core.examwork.service.ExamStudentService;
import cn.com.qmth.examcloud.core.examwork.service.bean.ExamInfo;
import cn.com.qmth.examcloud.core.examwork.service.bean.ExamSpecialSettingsInfo;
import cn.com.qmth.examcloud.task.api.DataSyncCloudService;
import cn.com.qmth.examcloud.task.api.request.SyncExamReq;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;

/**
 * 类注释
 *
 * @author WANGWEI
 * @date 2018年8月17日
 * @Copyright (c) 2018-? http://qmth.com.cn All Rights Reserved.
 */
@Service
public class ExamServiceImpl implements ExamService {
	@Autowired
	ExamRepo examRepo;

	@Autowired
	ExamStudentService examStudentService;

	@Autowired
	ExamSpecialSettingsRepo examSpecialSettingsRepo;

	@Autowired
	OrgCloudService orgCloudService;

	@Autowired
	ExamPropertyRepo examPropertyRepo;

	@Autowired
	ExamOrgPropertyRepo examOrgPropertyRepo;

	@Autowired
	DataSyncCloudService dataSyncCloudService;

	private static final String[] EXAM_ORG_SETTINGS_EXCEL_HEADER = new String[]{"学习中心ID", "学习中心代码",
			"学习中心名称", "是否可以考试(是/否)", "开始考试时间 yyyy-MM-dd hh:mm:ss", "结束考试时间 yyyy-MM-dd hh:mm:ss"};

	/*
	 * 实现
	 *
	 * @author WANGWEI
	 * 
	 * @see
	 * cn.com.qmth.examcloud.core.examwork.service.ExamService#saveExam(cn.com.
	 * qmth.examcloud.core.examwork.service.bean.ExamInfo)
	 */
	@Override
	public ExamEntity saveExam(ExamInfo examInfo, CURD curd) {

		Long rootOrgId = examInfo.getRootOrgId();
		String code = examInfo.getCode();
		String name = examInfo.getName();
		ExamType examType = examInfo.getExamType();

		if (null == examType) {
			throw new StatusException("001005", "考试类型不能为空");
		}

		if (StringUtils.isBlank(code)) {
			throw new StatusException("001005", "考试编码不能为空");
		}

		if (null == rootOrgId) {
			throw new StatusException("001005", "rootOrgId is null");
		}

		GetOrgReq getOrgReq = new GetOrgReq();
		getOrgReq.setOrgId(rootOrgId);
		GetOrgResp getOrgResp = orgCloudService.getOrg(getOrgReq);
		OrgBean rootOrg = getOrgResp.getOrg();

		Map<String, String> properties = examInfo.getProperties();
		Map<DynamicEnum, String> map = checkAndGetExamProperties(properties);

		ExamEntity exam = null;
		CURD realStatus = null;

		// 更新
		if (curd.equals(CURD.UPDATE)) {
			exam = examRepo.findByCodeAndRootOrgId(code, rootOrgId);
			if (null == exam) {
				throw new StatusException("002001", "code is wrong");
			}
			realStatus = CURD.UPDATE;
		}
		// 创建
		else if (curd.equals(CURD.CREATION)) {
			exam = new ExamEntity();
			realStatus = CURD.CREATION;
		}
		// (根据考试编码)新增或创建
		else if (curd.equals(CURD.CREATION_OR_UPDATE)) {
			exam = examRepo.findByCodeAndRootOrgId(code, rootOrgId);
			// 创建
			if (null == exam) {
				exam = new ExamEntity();
				realStatus = CURD.CREATION;
			}
			// 更新
			else {
				realStatus = CURD.UPDATE;
			}
		}

		if (realStatus.equals(CURD.CREATION)) {
			if (null == examInfo.getBeginTime()) {
				throw new StatusException("002006", "beginTime is null");
			}
			if (null == examInfo.getEndTime()) {
				throw new StatusException("002006", "endTime is null");
			}
			if (StringUtils.isBlank(name)) {
				throw new StatusException("002004", "name is blank");
			}

			ExamEntity examByCode = examRepo.findByCodeAndRootOrgId(code, rootOrgId);
			if (null != examByCode) {
				throw new StatusException("002005", "考试编码已存在");
			}

			ExamEntity examByName = examRepo.findByNameAndRootOrgId(name, rootOrgId);
			if (null != examByName) {
				throw new StatusException("002005", "考试名称已存在");
			}

			exam.setCode(code);
			exam.setRootOrgId(rootOrgId);
			exam.setExamType(examType);
			exam.setEnable(true);
		} else if (realStatus.equals(CURD.UPDATE)) {
			if (!exam.getRootOrgId().equals(rootOrgId)) {
				throw new StatusException("002003", "rootOrgId is wrong");
			}
			if (!exam.getExamType().equals(examType)) {
				throw new StatusException("002100", "examType is wrong");
			}
		}

		if (null != examInfo.getBeginTime()) {
			exam.setBeginTime(examInfo.getBeginTime());
		}
		if (null != examInfo.getEndTime()) {
			exam.setEndTime(examInfo.getEndTime());
		}
		if (null != examInfo.getDuration()) {
			exam.setDuration(examInfo.getDuration());
		}
		if (null != examInfo.getEnable()) {
			exam.setEnable(examInfo.getEnable());
		}
		if (null != examInfo.getExamTimes()) {
			exam.setExamTimes(examInfo.getExamTimes());
		}
		if (null != examInfo.getExamLimit()) {
			exam.setExamLimit(examInfo.getExamLimit());
		}
		if (StringUtils.isNotBlank(name)) {
			exam.setName(name);
		}
		if (StringUtils.isNotBlank(examInfo.getRemark())) {
			exam.setName(examInfo.getRemark());
		}

		ExamEntity saved = examRepo.saveAndFlush(exam);

		for (Entry<DynamicEnum, String> entry : map.entrySet()) {
			DynamicEnum de = entry.getKey();
			String value = entry.getValue();
			ExamPropertyEntity entity = examPropertyRepo.findByExamIdAndKeyId(saved.getId(),
					de.getId());
			if (null == entity) {
				entity = new ExamPropertyEntity();
				entity.setExamId(saved.getId());
				entity.setKeyId(de.getId());
			}
			entity.setValue(value);

			examPropertyRepo.save(entity);
		}

		// 考试设置混入特殊设置
		ExamSpecialSettingsInfo examSpecialInfo = new ExamSpecialSettingsInfo();
		examSpecialInfo.setBeginTime(saved.getBeginTime());
		examSpecialInfo.setEndTime(saved.getEndTime());
		examSpecialInfo.setExamId(saved.getId());
		examSpecialInfo.setExamLimit(saved.getExamLimit());
		examSpecialInfo.setRootOrgId(saved.getRootOrgId());
		this.saveExamSpecialSettings(examSpecialInfo);

		examSpecialSettingsRepo.updateExamEnableByExamId(saved.getId(), saved.getEnable());

		SyncExamReq req = new SyncExamReq();
		req.setId(saved.getId());
		req.setBeginTime(saved.getBeginTime());
		req.setDuration(saved.getDuration());
		req.setEnable(saved.getEnable());
		req.setEndTime(saved.getEndTime());
		req.setExamTimes(saved.getExamTimes());
		req.setExamType(saved.getExamType().name());
		req.setName(saved.getName());
		req.setRemark(saved.getRemark());
		req.setRootOrgId(saved.getRootOrgId());
		req.setRootOrgName(rootOrg.getName());
		req.setSyncType("update");
		dataSyncCloudService.syncExam(req);

		return saved;
	}

	/**
	 * 方法注释
	 *
	 * @author WANGWEI
	 * @param properties
	 * @return
	 */
	private Map<DynamicEnum, String> checkAndGetExamProperties(Map<String, String> properties) {
		DynamicEnumManager manager = ExamProperty.getDynamicEnumManager();

		Map<DynamicEnum, String> map = Maps.newHashMap();
		if (null == properties) {
			return map;
		}

		for (Entry<String, String> entry : properties.entrySet()) {
			String key = entry.getKey();
			String value = entry.getValue();
			DynamicEnum de = null;
			try {
				de = manager.getByName(key);
			} catch (Exception e) {
				throw new StatusException("001004", "考试属性错误. key=" + key);
			}
			if (!de.isLegal(value)) {
				throw new StatusException("001200", "考试属性值非法. key=" + key);
			}
			map.put(de, value);
		}

		String beforeExamRemark = properties.get("BEFORE_EXAM_REMARK");
		if (null != beforeExamRemark && beforeExamRemark.length() > 250000) {
			throw new StatusException("001002", "考前说明内容过大");
		}

		String afterExamRemark = properties.get("AFTER_EXAM_REMARK");
		if (null != afterExamRemark && afterExamRemark.length() > 250000) {
			throw new StatusException("001002", "考后说明内容过大");
		}

		String cheatingRemark = properties.get("CHEATING_REMARK");
		if (null != cheatingRemark && cheatingRemark.length() > 250000) {
			throw new StatusException("001002", "作弊说明内容过大");
		}

		return map;
	}

	/*
	 * 实现
	 *
	 * @author WANGWEI
	 * 
	 * @see
	 * cn.com.qmth.examcloud.core.examwork.service.ExamService#saveExamOrg(cn.
	 * com.qmth.examcloud.core.examwork.service.bean.ExamOrgInfo)
	 */
	@Override
	public ExamSpecialSettingsEntity saveExamSpecialSettings(
			ExamSpecialSettingsInfo examSpecialInfo) {
		ExamSpecialSettingsEntity examOrgEntity = null;

		Long examId = examSpecialInfo.getExamId();
		Date beginTime = examSpecialInfo.getBeginTime();
		Date endTime = examSpecialInfo.getEndTime();

		if (!new Boolean(null == beginTime).equals(new Boolean(null == endTime))) {
			throw new StatusException("001101", "beginTime & endTime  wrong");
		}

		ExamEntity examEntity = GlobalHelper.getPresentEntity(examRepo, examId, ExamEntity.class);

		if (null == examSpecialInfo.getId()) {
			examOrgEntity = examSpecialSettingsRepo.findByExamIdAndOrgIdAndCourseIdIsNull(
					examSpecialInfo.getExamId(), examSpecialInfo.getOrgId());
			if (null == examOrgEntity) {
				examOrgEntity = new ExamSpecialSettingsEntity();
			}
		} else {
			examOrgEntity = GlobalHelper.getEntity(examSpecialSettingsRepo, examSpecialInfo.getId(),
					ExamSpecialSettingsEntity.class);
			if (null == examOrgEntity) {
				throw new StatusException("001101", "id is wrong");
			}
		}

		examOrgEntity.setBeginTime(examSpecialInfo.getBeginTime());
		examOrgEntity.setEndTime(examSpecialInfo.getEndTime());
		examOrgEntity.setRootOrgId(examSpecialInfo.getRootOrgId());
		examOrgEntity.setExamId(examSpecialInfo.getExamId());
		examOrgEntity.setOrgId(examSpecialInfo.getOrgId());
		examOrgEntity.setExamType(examEntity.getExamType());
		examOrgEntity.setExamEnable(examEntity.getEnable());
		Boolean examLimit = examSpecialInfo.getExamLimit();
		examLimit = null == examLimit ? false : examLimit;
		examOrgEntity.setExamLimit(examLimit);

		Map<DynamicEnum, String> map = checkAndGetExamProperties(examSpecialInfo.getProperties());

		ExamSpecialSettingsEntity saved = examSpecialSettingsRepo.save(examOrgEntity);

		for (Entry<DynamicEnum, String> entry : map.entrySet()) {
			DynamicEnum de = entry.getKey();
			String value = entry.getValue();
			if (StringUtils.isBlank(value)) {
				value = null;
			} else {
				value = value.trim();
			}
			ExamOrgPropertyEntity entity = examOrgPropertyRepo
					.findByexamIdAndOrgIdAndKeyId(saved.getExamId(), saved.getOrgId(), de.getId());
			if (null == entity) {
				entity = new ExamOrgPropertyEntity();
				entity.setExamId(saved.getExamId());
				entity.setOrgId(saved.getOrgId());
				entity.setKeyId(de.getId());
			}
			entity.setValue(value);

			examOrgPropertyRepo.save(entity);
		}

		return saved;
	}

	@Override
	public String getOrgProperty(Long examId, Long orgId, String key) {
		DynamicEnumManager manager = ExamProperty.getDynamicEnumManager();
		DynamicEnum de = manager.getByName(key);

		if (null != orgId) {
			ExamOrgPropertyEntity examOrgPropertyEntity = examOrgPropertyRepo
					.findByexamIdAndOrgIdAndKeyId(examId, orgId, de.getId());
			if (null != examOrgPropertyEntity) {
				return examOrgPropertyEntity.getValue();
			}
		}

		ExamPropertyEntity examPropertyEntity = examPropertyRepo.findByExamIdAndKeyId(examId,
				de.getId());
		if (null == examPropertyEntity) {
			return null;
		}

		return examPropertyEntity.getValue();
	}

	@Override
	public List<Map<String, Object>> importExamSpecialSettings(Long examId, File file) {

		ExamEntity examEntity = GlobalHelper.getEntity(examRepo, examId, ExamEntity.class);
		List<String[]> lineList = null;
		try {
			lineList = ExcelReader.readSheetBySax(PathUtil.getCanonicalPath(file), 1, 6);
		} catch (Exception e) {
			throw new StatusException("200110", "Excel 解析失败");
		}

		if (CollectionUtils.isEmpty(lineList)) {
			throw new StatusException("200111", "Excel无内容");
		}

		if (10001 < lineList.size()) {
			throw new StatusException("200112", "数据行数不能超过10000");
		}

		List<Map<String, Object>> failRecords = Collections
				.synchronizedList(new ArrayList<Map<String, Object>>());
		List<ExamSpecialSettingsEntity> specialSettingsList = Lists.newArrayList();

		for (int i = 0; i < lineList.size(); i++) {
			String[] line = lineList.get(i);
			if (0 == i) {
				if (headerError(line)) {
					throw new StatusException("100111", "Excel表头错误");
				}
				continue;
			}

			boolean hasError = false;
			StringBuilder msg = new StringBuilder();

			ExamSpecialSettingsEntity specialSettings = new ExamSpecialSettingsEntity();
			specialSettings.setRootOrgId(examEntity.getRootOrgId());
			specialSettings.setExamId(examId);

			String orgId = trimAndNullIfBlank(line[0]);
			if (StringUtils.isBlank(orgId)) {
				msg.append("  学习中心ID不能为空");
				hasError = true;
			} else {
				try {
					long orgIdLong = Long.parseLong(orgId);

					GetOrgReq getOrgReq = new GetOrgReq();
					getOrgReq.setOrgId(orgIdLong);
					GetOrgResp getOrgResp = orgCloudService.getOrg(getOrgReq);
					OrgBean orgBean = getOrgResp.getOrg();

					if (null == orgBean.getParentId()
							|| !orgBean.getRootId().equals(examEntity.getRootOrgId())) {
						msg.append("  学习中心ID非法");
						hasError = true;
					} else {
						specialSettings.setOrgId(orgIdLong);
					}
				} catch (NumberFormatException e) {
					msg.append("  学习中心ID必须为整数");
					hasError = true;
				}
			}

			String examLimit = trimAndNullIfBlank(line[3]);
			if (StringUtils.isNotBlank(examLimit)) {
				if (examLimit.equals("是")) {
					specialSettings.setExamLimit(false);
				} else if (examLimit.equals("否")) {
					specialSettings.setExamLimit(true);
				} else {
					msg.append("  是否可以开始考试必须为['是','否',空]");
					hasError = true;
				}
			} else {
				specialSettings.setExamLimit(false);
			}

			String beginTime = trimAndNullIfBlank(line[4]);
			if (StringUtils.isNotBlank(beginTime)) {
				try {
					Date beginDate = DateUtil.parse(beginTime, DatePatterns.CHINA_DEFAULT);
					specialSettings.setBeginTime(beginDate);
				} catch (Exception e) {
					msg.append("  开始考试时间格式错误. 正确格式为 " + DatePatterns.CHINA_DEFAULT);
					hasError = true;
				}
			}
			String endTime = trimAndNullIfBlank(line[5]);
			if (StringUtils.isNotBlank(endTime)) {
				try {
					Date endDate = DateUtil.parse(endTime, DatePatterns.CHINA_DEFAULT);
					specialSettings.setEndTime(endDate);
				} catch (Exception e) {
					msg.append("  结束考试时间格式错误. 正确格式为 " + DatePatterns.CHINA_DEFAULT);
					hasError = true;
				}
			}

			if (hasError) {
				failRecords.add(newError(i + 1, msg.toString()));
			} else {
				specialSettingsList.add(specialSettings);
			}

		}

		if (CollectionUtils.isNotEmpty(failRecords)) {
			return failRecords;
		}

		for (ExamSpecialSettingsEntity cur : specialSettingsList) {
			ExamSpecialSettingsEntity query = examSpecialSettingsRepo
					.findByExamIdAndOrgIdAndCourseIdIsNull(examId, cur.getOrgId());

			if (null != query) {
				query.setExamType(examEntity.getExamType());
				query.setExamEnable(examEntity.getEnable());
				query.setExamLimit(cur.getExamLimit());
				query.setBeginTime(cur.getBeginTime());
				query.setEndTime(cur.getEndTime());
				examSpecialSettingsRepo.save(query);
			} else {
				cur.setExamType(examEntity.getExamType());
				cur.setExamEnable(examEntity.getEnable());
				examSpecialSettingsRepo.save(cur);
			}
		}

		return failRecords;

	}

	private Map<String, Object> newError(int lineNum, String msg) {
		Map<String, Object> map = Maps.newHashMap();
		map.put("lineNum", lineNum);
		map.put("msg", msg);
		return map;
	}

	private String trimAndNullIfBlank(String s) {
		if (StringUtils.isBlank(s)) {
			return null;
		}
		return s.trim();
	}

	/**
	 * 方法注释
	 *
	 * @author WANGWEI
	 * @param header
	 */
	private boolean headerError(String[] header) {
		for (int i = 0; i < EXAM_ORG_SETTINGS_EXCEL_HEADER.length; i++) {
			if (null == header[i]) {
				return true;
			}
			if (!EXAM_ORG_SETTINGS_EXCEL_HEADER[i].equals(header[i].trim())) {
				return true;
			}
		}
		return false;
	}

}
